@@ -9,8 +9,7 @@
 #include "libchash.h"
 
 struct bucket_t {
-    const char *node_name;
-    size_t name_len;
+    uint32_t node_idx;
     uint32_t point;
 } bucket_t;
 
@@ -24,8 +23,8 @@ struct chash_t {
 
 static int cmpbucket(const void *a, const void *b)
 {
-    struct bucket_t *b1 = (struct bucket_t *) a;
-    struct bucket_t *b2 = (struct bucket_t *) b;
+    const struct bucket_t *b1 = (const struct bucket_t *) a;
+    const struct bucket_t *b2 = (const struct bucket_t *) b;
 
     if (b1->point < b2->point) {
 	return -1;
@@ -39,7 +38,7 @@ static int cmpbucket(const void *a, const void *b)
 }
 
 /* from leveldb, a murmur-lite */
-static uint32_t leveldb_bloom_hash(unsigned char *b, size_t len)
+static uint32_t leveldb_bloom_hash(const unsigned char *b, size_t len)
 {
     const uint32_t seed = 0xbc9f1d34;
     const uint32_t m = 0xc6a4a793;
@@ -66,26 +65,38 @@ static uint32_t leveldb_bloom_hash(unsigned char *b, size_t len)
     return h;
 }
 
-struct chash_t *chash_create(const char **node_names, size_t * name_lens,
+struct chash_t *chash_create(const char **node_names, size_t *name_lens,
 			     size_t num_names, size_t replicas)
 {
     struct chash_t *chash;
-
-    struct bucket_t *blist =
-	(struct bucket_t *) malloc(sizeof(bucket_t) * num_names * replicas);
-    char **nlist = (char **) malloc(sizeof(char *) * num_names);
-    size_t *lens = (size_t *) malloc(sizeof(size_t) * num_names);
-    size_t n, r, len, len1, len2, bidx = 0;
-
+    struct bucket_t *blist;
+    char **nlist;
+    size_t *lens;
+    size_t i, n, r, len, len1, len2, bidx = 0;
     char buffer[256];
 
+    blist = (struct bucket_t *) malloc(sizeof(bucket_t) * num_names * replicas);
+    if (blist == NULL) {
+	goto ERROR_BLIST;
+    }
+    nlist = (char **) malloc(sizeof(char *) * num_names);
+    if (nlist == NULL) {
+	goto ERROR_NLIST;
+    }
+    lens = (size_t *) malloc(sizeof(size_t) * num_names);
+    if (lens == NULL) {
+	goto ERROR_LENS;
+    }
+
     for (n = 0; n < num_names; n++) {
 	nlist[n] = (char *) malloc(sizeof(char) * name_lens[n]);
+	if (nlist[n] == NULL) {
+	    goto ERROR_NLIST_MEMBER;
+	}
 	lens[n] = name_lens[n];
 	memcpy(nlist[n], node_names[n], lens[n]);
 	for (r = 0; r < replicas; r++) {
-	    blist[bidx].node_name = nlist[n];
-	    blist[bidx].name_len = lens[n];
+	    blist[bidx].node_idx = n;
 	    len1 = snprintf(buffer, sizeof(buffer), "%lu", (unsigned long) r);
 	    if ((sizeof(buffer) - len1) < name_lens[n]) {
 		len2 = sizeof(buffer) - len1;
@@ -99,7 +110,7 @@ struct chash_t *chash_create(const char **node_names, size_t * name_lens,
 		fprintf(stderr, "Node name truncated to: %s\n", buffer);
 	    }
 	    blist[bidx].point =
-		leveldb_bloom_hash((unsigned char *) buffer, len);
+		leveldb_bloom_hash((const unsigned char *) buffer, len);
 	    bidx++;
 	}
     }
@@ -107,6 +118,9 @@ struct chash_t *chash_create(const char **node_names, size_t * name_lens,
     qsort(blist, bidx, sizeof(struct bucket_t), cmpbucket);
 
     chash = malloc(sizeof(chash_t));
+    if (chash == NULL) {
+	goto ERROR_CHASH;
+    }
     chash->blist = blist;
     chash->nbuckets = bidx;
     chash->node_names = nlist;
@@ -114,14 +128,30 @@ struct chash_t *chash_create(const char **node_names, size_t * name_lens,
     chash->num_names = num_names;
 
     return chash;
+
+  ERROR_CHASH:
+  ERROR_NLIST_MEMBER:
+    /* We know that nlist[n] is the first member not allocated,
+       thus free every nlist[i] up to n. */
+    for (i = 0; i < n; i++) {
+	free(nlist[n]);
+    }
+    free(lens);
+  ERROR_LENS:
+    free(nlist);
+  ERROR_NLIST:
+    free(blist);
+  ERROR_BLIST:
+    return NULL;
 }
 
 void chash_lookup(struct chash_t *chash, const char *key, size_t len,
-		  const char **node_name, size_t * name_len)
+		  const char **node_name, size_t *name_len)
 {
-    uint32_t point = leveldb_bloom_hash((unsigned char *) key, len);
+    uint32_t point = leveldb_bloom_hash((const unsigned char *) key, len);
 
     uint32_t low = 0, high = chash->nbuckets;
+    uint32_t node_idx;
 
     /* binary search through blist */
     while (low < high) {
@@ -137,8 +167,9 @@ void chash_lookup(struct chash_t *chash, const char *key, size_t len,
 	low = 0;
     }
 
-    *node_name = chash->blist[low].node_name;
-    *name_len = chash->blist[low].name_len;
+    node_idx = chash->blist[low].node_idx;
+    *node_name = chash->node_names[node_idx];
+    *name_len = chash->name_lens[node_idx];
 }
 
 void chash_free(struct chash_t *chash)
